@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #include <stdio.h>
@@ -38,46 +38,6 @@
 #include "libcryptsetup.h"
 #include "internal.h"
 
-static char *error=NULL;
-
-__attribute__((format(printf, 1, 0)))
-void set_error_va(const char *fmt, va_list va)
-{
-	int r;
-
-	if(error) {
-		free(error);
-		error = NULL;
-	}
-
-	if(!fmt) return;
-
-	r = vasprintf(&error, fmt, va);
-	if (r < 0) {
-		free(error);
-		error = NULL;
-		return;
-	}
-
-	if (r && error[r - 1] == '\n')
-		error[r - 1] = '\0';
-}
-
-__attribute__((format(printf, 1, 2)))
-void set_error(const char *fmt, ...)
-{
-	va_list va;
-
-	va_start(va, fmt);
-	set_error_va(fmt, va);
-	va_end(va);
-}
-
-const char *get_error(void)
-{
-	return error;
-}
-
 static int get_alignment(int fd)
 {
 	int alignment = DEFAULT_MEM_ALIGNMENT;
@@ -336,10 +296,23 @@ int device_ready(struct crypt_device *cd, const char *device, int mode)
 	return r;
 }
 
-int get_device_infos(const char *device,
-		     int open_exclusive,
-		     int *readonly,
-		     uint64_t *size)
+int device_size(const char *device, uint64_t *size)
+{
+	int devfd, r = 0;
+
+	devfd = open(device, O_RDONLY);
+	if(devfd == -1)
+		return -EINVAL;
+
+	if (ioctl(devfd, BLKGETSIZE64, size) < 0)
+		r = -EINVAL;
+
+	close(devfd);
+	return r;
+}
+
+static int get_device_infos(const char *device, enum devcheck device_check,
+			    int *readonly, uint64_t *size)
 {
 	struct stat st;
 	unsigned long size_small;
@@ -353,7 +326,7 @@ int get_device_infos(const char *device,
 		return -EINVAL;
 
 	/* never wipe header on mounted device */
-	if (open_exclusive && S_ISBLK(st.st_mode))
+	if (device_check == DEV_EXCL && S_ISBLK(st.st_mode))
 		flags |= O_EXCL;
 
 	/* Try to open read-write to check whether it is a read-only device */
@@ -363,7 +336,7 @@ int get_device_infos(const char *device,
 		fd = open(device, O_RDONLY | flags);
 	}
 
-	if (fd == -1 && open_exclusive && errno == EBUSY)
+	if (fd == -1 && device_check == DEV_EXCL && errno == EBUSY)
 		return -EBUSY;
 
 	if (fd == -1)
@@ -396,10 +369,10 @@ out:
 
 int device_check_and_adjust(struct crypt_device *cd,
 			    const char *device,
-			    int open_exclusive,
+			    enum devcheck device_check,
 			    uint64_t *size,
 			    uint64_t *offset,
-			    int *read_only)
+			    uint32_t *flags)
 {
 	int r, real_readonly;
 	uint64_t real_size;
@@ -407,7 +380,7 @@ int device_check_and_adjust(struct crypt_device *cd,
 	if (!device)
 		return -ENOTBLK;
 
-	r = get_device_infos(device, open_exclusive, &real_readonly, &real_size);
+	r = get_device_infos(device, device_check, &real_readonly, &real_size);
 	if (r < 0) {
 		if (r == -EBUSY)
 			log_err(cd, _("Cannot use device %s which is in use "
@@ -419,62 +392,49 @@ int device_check_and_adjust(struct crypt_device *cd,
 		return r;
 	}
 
+	if (*offset >= real_size) {
+		log_err(cd, _("Requested offset is beyond real size of device %s.\n"),
+			device);
+		return -EINVAL;
+	}
+
 	if (!*size) {
 		*size = real_size;
 		if (!*size) {
 			log_err(cd, _("Device %s has zero size.\n"), device);
 			return -ENOTBLK;
 		}
-		if (*size < *offset) {
-			log_err(cd, _("Device %s is too small.\n"), device);
-			return -EINVAL;
-		}
 		*size -= *offset;
 	}
 
+	/* in case of size is set by parameter */
+	if ((real_size - *offset) < *size) {
+		log_dbg("Device %s: offset = %" PRIu64 " requested size = %" PRIu64
+			", backing device size = %" PRIu64,
+			device, *offset, *size, real_size);
+		log_err(cd, _("Device %s is too small.\n"), device);
+		return -EINVAL;
+	}
+
+	if (device_check == DEV_SHARED) {
+		log_dbg("Checking crypt segments for device %s.", device);
+		r = crypt_sysfs_check_crypt_segment(device, *offset, *size);
+		if (r < 0) {
+			log_err(cd, _("Cannot use device %s (crypt segments "
+				    "overlaps or in use by another device).\n"),
+				    device);
+			return r;
+		}
+	}
+
 	if (real_readonly)
-		*read_only = 1;
+		*flags |= CRYPT_ACTIVATE_READONLY;
 
 	log_dbg("Calculated device size is %" PRIu64 " sectors (%s), offset %" PRIu64 ".",
-		*size, *read_only ? "RO" : "RW", *offset);
+		*size, real_readonly ? "RO" : "RW", *offset);
 	return 0;
 }
 
-int wipe_device_header(const char *device, int sectors)
-{
-	struct stat st;
-	char *buffer;
-	int size = sectors * SECTOR_SIZE;
-	int r = -1;
-	int devfd;
-	int flags = O_RDWR | O_DIRECT | O_SYNC;
-
-	if (stat(device, &st) < 0)
-		return -EINVAL;
-
-	/* never wipe header on mounted device */
-	if (S_ISBLK(st.st_mode))
-		flags |= O_EXCL;
-
-	devfd = open(device, flags);
-	if(devfd == -1)
-		return errno == EBUSY ? -EBUSY : -EINVAL;
-
-	buffer = malloc(size);
-	if (!buffer) {
-		close(devfd);
-		return -ENOMEM;
-	}
-	memset(buffer, 0, size);
-
-	r = write_blockwise(devfd, buffer, size) < size ? -EIO : 0;
-
-	free(buffer);
-	close(devfd);
-
-	return r;
-}
-
 /* MEMLOCK */
 #define DEFAULT_PROCESS_PRIORITY -18
 
